import rospy
from threading import Lock
from CMap2D import CMap2D
# if this is after remove_python2..., somehow python2 and python3 mismatch leads to non-connecting topics
from nav_msgs.msg import OccupancyGrid

def remove_python2_entries_from_sys_path():
    """ sourcing ros means adding its python2 paths to sys.path. This creates incompatibilities when
    importing tf, so we remove them """
    import sys
    if sys.version[0] == str(3):
        new_path = []
        for p in sys.path:
            if "python2" in p:
                print("REMOVING python2 entry from sys.path: {}".format(p))
                continue
            new_path.append(p)
        sys.path = new_path


remove_python2_entries_from_sys_path()
import tf

def default_refmap_update_callback(self_):
    pass

""" so many nodes end up implementing this functionality, it made sense to create
a class for it.

This class loads and stores the reference map (for global planning, for example),
and keeps track of the robot's pose in the map

with refmap_update_callback, the user can specify a task to perform each time the map is updated
"""
class ReferenceMapAndLocalizationManager(object):
    """ If a reference map is provided and a tf exists,
    keeps track of the tf for given frame in the reference map frame """
    def __init__(self, map_folder, map_filename, reference_map_frame, frame_to_track, refmap_update_callback=default_refmap_update_callback):
        self.tf_frame_in_refmap = None
        self.map_ = None
        self.lock = Lock() # for avoiding race conditions
        self.refmap_is_dynamic = False
        # update callback
        self.refmap_update_callback = refmap_update_callback
        # get frame info
        self.kRefMapFrame = reference_map_frame
        self.kFrame = frame_to_track
        # Load map
        if map_folder == "rostopic":
            self.refmap_is_dynamic = True
            rospy.logwarn("Getting reference map from topic '{}'".format(map_filename))
            rospy.Subscriber(map_filename, OccupancyGrid, self.map_callback, queue_size=1)
        else:
            # loads map based on ros params
            folder = map_folder
            filename =  map_filename
            try:
                self.map_ = CMap2D(folder, filename)
            except IOError as e:
                rospy.logwarn(rospy.get_namespace())
                rospy.logwarn("Failed to load reference map. Make sure {}.yaml and {}.pgm"
                       " are in the {} folder.".format(filename, filename, folder))
                rospy.logwarn("Disabling. No global localization or reference map will be available.")
                return
            self.refmap_update_callback(self)
        # launch callbacks
        self.tf_listener = tf.TransformListener()
        rospy.Timer(rospy.Duration(0.01), self.tf_callback)
        self.has_last_failed = False

    def tf_callback(self, event=None):
        try:
             self.tf_frame_in_refmap = self.tf_listener.lookupTransform(self.kRefMapFrame, self.kFrame, rospy.Time(0))
             if self.has_last_failed:
                 rospy.logwarn("refmap tf found.")
                 self.has_last_failed = False
        except (tf.LookupException, tf.ConnectivityException, tf.ExtrapolationException) as e:
            rospy.logwarn_throttle(10., e)
            self.has_last_failed = True
            return

    def map_callback(self, msg):
        with self.lock:
            self.map_ = CMap2D()
            self.map_.from_msg(msg)
            self.refmap_update_callback(self)




import numpy as np

def Pose2D(tf_msg):
    import tf
    return np.array([tf_msg[0][0], tf_msg[0][1], 
        tf.transformations.euler_from_quaternion(tf_msg[1])[2]])

def posemsg_to_pose2d(posemsg):
    import tf
    quaternion = (
        posemsg.orientation.x,
        posemsg.orientation.y,
        posemsg.orientation.z,
        posemsg.orientation.w,
    )
    xytheta = np.array([0., 0., 0.])
    xytheta[0] = posemsg.position.x
    xytheta[1] = posemsg.position.y
    xytheta[2] = tf.transformations.euler_from_quaternion(quaternion)[2]
    return xytheta

def apply_tf(x, pose2d):
    """ Applies transform to x
    x is a list of x y points in frame B
    pose2d is the transform AT_B, (frame B in frame A)
    result is x in frame A

    Parameters
    ----------
        pose2d : np.array
            Array of shape (3,), x, y, theta components of the tf of frame B in frame A
        x: np.array
            Array of shape (N, 2), x, y coordinates of N points in frame B
    Returns
    -------
        result: np.array
            Array of shape (N, 2), x, y coordinates of N points in frame A

    Note
    ----
    pose2d is under the form [x, y, theta], where 
    x and y are coordinates of frame B in frame A, 
    and theta is the orientation of frame B in frame A.

    Example
    -------
    >>> BX = np.array([[0, 0], [1, 0]])
    >>> AT_B = np.array([2, 1, np.pi/2])
    >>> apply_tf(BX, AT_B)
    array([[ 2.,  1.],
           [ 2.,  2.]])
    """
    return rotate(x, pose2d[2]) + np.array(pose2d[:2])

def apply_tf_to_pose(pose, pose2d):
    # same as apply_tf but assumes pose is x y theta instead of x y
    xy = rotate(np.array([pose[:2]]), pose2d[2])[0] + np.array(pose2d[:2])
    th = pose[2] + pose2d[2]
    return np.array([xy[0], xy[1], th])

def apply_tf_to_poses(poses, pose2d):
    """
    Parameters
    ----------
        poses: np.array
            Array of shape (N, 3), x, y, theta components of tfs of frame C in frame B)
        pose2d : np.array
            Array of shape (3,), x, y, theta components of the tf of frame B in frame A
    Returns
    -------
        result: np.array
            Array of shape (N, 3), x, y theta components of the tf of frame C in frame A
    """
    result = np.zeros_like(poses)
    xy = rotate(poses[:, :2], pose2d[2]) + pose2d[:2][None, :]
    th = poses[:, 2] + pose2d[2]
    result[:, :2] = xy
    result[:, 2] = th
    return result

def apply_tf_to_vel(vel, pose2d):
    # same as apply_tf but assumes vel is xdot ydot thetadot instead of x y
    # for xdot ydot frame transformation applies, 
    # but thetadot is invariant due to frames being fixed.
    xy = rotate(vel[...,:2], pose2d[2])
    th = vel[...,2]
    xyth = np.zeros_like(vel)
    xyth[...,:2] = xy
    xyth[...,2] = th
    return xyth

def old_apply_tf_to_vel(vel, pose2d):
    xy = rotate(np.array([vel[...,:2]]), pose2d[2])[0]
    th = vel[...,2]
    return np.array([xy[0], xy[1], th])

def rotate(x, th):
    rotmat = np.array([
        [np.cos(th), -np.sin(th)],
        [np.sin(th), np.cos(th)],
    ])
    return np.matmul(rotmat, x.T).T

def inverse_pose2d(pose2d):
    inv_th = -pose2d[2] # theta
    inv_xy = rotate(np.array([-pose2d[:2]]), inv_th)[0]
    return np.array([inv_xy[0], inv_xy[1], inv_th])


if __name__ == "__main__":
    import doctest

    doctest.testmod()
